/* 
 *  ufdbguardd.c  -  multithreaded ufdbGuard daemon
 *
 *  $Id: ufdbguardd.c,v 1.29 2006/12/07 10:10:28 root Exp root $
 */

#include "sg.h"
#include "ufdb.h"
#include "ufdbchkport.h"
#include "httpsQueue.h"

#include <stdio.h>
#include <errno.h>
#include <pthread.h>
#include <signal.h>
#include <libgen.h>
#include <unistd.h>
#include <sys/utsname.h>
#include <sys/syscall.h>
#ifdef __linux__
#include <linux/unistd.h>
#endif
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/tcp.h>

#ifdef __linux__
#include <execinfo.h>
#endif

#define FULL 1
#define EMPTY 0

pthread_mutex_t sighup_mutex = PTHREAD_MUTEX_INITIALIZER;

#define UFDB_MAX_WORKERS	UFDB_MAX_THREADS
static int     n_workers = 33;

static int     my_argc;
static char ** my_argv;
static char ** my_envp;
static char *  configFile = NULL;
static struct timeval  start_time;

#define UFDBGUARDD_PID_FILE	"/var/tmp/ufdbguardd.pid"
char * globalPidFile = UFDBGUARDD_PID_FILE;

extern int globalPid;

extern int globalFatalError;
extern int globalDebugRedirect;
extern int globalDebugTimeDelta;
extern int globalDebugRedirect;
extern int sig_other;
extern char ** globalArgv;
extern char ** globalEnvp;

extern int analyseUncategorised;
extern int httpsConnectionCacheSize;

extern struct UFDBmemTable knownDb;

static int        portNumCmdLine = -1;
extern int        portNum;

static int        runAsDaemon = 1;
static int        noRedirects = 0;
static int        parseOnly = 0;
static pthread_t  HUPhandler;
static pthread_t  sockethandler;
static pthread_t  workers[UFDB_MAX_WORKERS];
static pthread_t  httpsthr[UFDB_NUM_HTTPS_VERIFIERS];

volatile int UFDBreconfig = 1;

volatile unsigned long UFDBlookupCounter = 0;
volatile unsigned long UFDBlookupHttpsCounter = 0;
volatile unsigned long UFDBblockCounter = 0;
volatile unsigned long UFDBtestBlockCounter = 0;
volatile unsigned long UFDBuncategorisedCounter = 0;
volatile unsigned long UFDBuploadSeqNo = 0;

static void saveUncategorisedURLs( void );
static void * worker_main( void * ptr );


static void usage( char option )
{
   if (option != '\0')
      fprintf( stderr, "unknown option '-%c'\n", option );

   fprintf( stderr, "usage: ufdbguardd [-A] [-d] [-v] [-r] [-T] [-h] [-w 8-64] [-p port] [-c <configfile>]\n" );
   fprintf( stderr, "-v    print version\n" );
   fprintf( stderr, "-c F  use configuration file F\n" );
   fprintf( stderr, "-T    test mode; do not block, only log which URLs might be blocked\n" );
   fprintf( stderr, "-w N  use N worker threads (default: 32)\n" );
   fprintf( stderr, "-p N  use TCP port N to listen on\n" );
   fprintf( stderr, "-d    extra debug information in log file\n" );
   fprintf( stderr, "-r    log all redirections\n" );
   fprintf( stderr, "-N    do not analyse uncategorised URLs\n" );
   fprintf( stderr, "-h    print this help text\n" );

   if (option != '\0')
      exit( 1 );
   else
      exit( 0 );
}


static void writePidFile( void )
{
   FILE * fp;

   unlink( globalPidFile );
   fp = fopen( globalPidFile, "w" );
   if (fp == NULL)
      ufdbLogError( "cannot write to PID file %s", globalPidFile );
   else
   {
      fprintf( fp, "%d\n", globalPid );
      fclose( fp );
   }
}


static void removePidFile( void )
{
   unlink( globalPidFile );
}


/* 
 * queue.c
 * 
 * communication between the worker and other threads is done via
 * RW-locks and data is passed via a queue.
 *
 * Since this queue is a FIFO we could use a tail queue. But we don't
 * like the malloc() overhead, so therefore we use a static array.
 *
 * RCS $Id: ufdbguardd.c,v 1.29 2006/12/07 10:10:28 root Exp root $
 */

#define UFDB_MAX_FD_QUEUE   UFDB_MAX_THREADS


static struct {
   int    fd;
} q[UFDB_MAX_FD_QUEUE];

static int    ihead;                    /* array index for the head */
static int    itail;                    /* array index for the tail */
static int    n_queued;
static pthread_mutex_t lock;
static pthread_cond_t  empty;


static void initFdQueue( void )
{
   ihead = 0;
   itail = 0;
   n_queued = 0;
   pthread_mutex_init( &lock, NULL );
   pthread_cond_init( &empty, NULL );
}


/*
 * enqueue a new fd.
 */
static void putFdQueue( int fd )
{
   pthread_mutex_lock( &lock );

   if (n_queued < UFDB_MAX_FD_QUEUE)
   {
      /*
       * insert it at the tail
       */
      q[itail].fd = fd;

      itail = (itail + 1) % UFDB_MAX_FD_QUEUE;
      n_queued++;

      /*
       * signal anyone who is waiting
       */
      pthread_cond_signal( &empty );

      /*
       * leave the critical section
       */
      pthread_mutex_unlock( &lock );
   } 
   else    
   {
      ufdbLogFatalError( "FATAL ERROR: connection queue is full." );
      ufdbLogFatalError( "There are too many ufdbgclient processes !!" );
      /* there is no other option than to close the connection */
      /* TO-DO: maybe: close all queued fd's ? */
      close( fd );
   }
}


/*
 * get a fd where there is new content.
 */
static void getFdQueue( int * fd )
{
   pthread_mutex_lock( &lock );

   while (1)
   {
      /*
       * if there are jobs in the queue
       */
      if (n_queued > 0)
      {
         /*
          * get the one at the head
          */
         *fd = q[ihead].fd;
         ihead = (ihead + 1) % UFDB_MAX_FD_QUEUE;
         n_queued--;

         /*
          * leave the critical section
          */
         pthread_mutex_unlock( &lock );

	 return;
      } 
      else   /* otherwise wait until there are fds available */ 
      {
         pthread_cond_wait( &empty, &lock );
         /*
          * when I'm here, I've been signaled because there
          * are jobs in the queue.  Go try and get one.
          */
      }
   }
}


static void daemon_accept_connections( int s, struct sockaddr * addr )
{
   int newfd;
   int size;
   int sock_parm;
   struct timeval tv;
   struct linger  linger;

   /*
    * Allow that this thread can be killed without delays at any time.
    */
   pthread_setcancelstate( PTHREAD_CANCEL_ENABLE, NULL );
   pthread_setcanceltype( PTHREAD_CANCEL_ASYNCHRONOUS, NULL );

   size = sizeof(struct sockaddr);
   while (!sig_other)
   {
      errno = 0;
      newfd = accept( s, addr, (socklen_t *) &size );
      if (newfd < 0)
      {
	 if (errno == EINTR)
	 {
	    continue;
	 }
	 ufdbLogError( "SRV: accept on socket failed: %s", strerror(errno) );
	 return;
      }
      linger.l_onoff = 1;
      linger.l_linger = 2;
      setsockopt( newfd, SOL_SOCKET, SO_LINGER, (void *) &linger, sizeof(linger) );

      sock_parm = 1;
      setsockopt( newfd, IPPROTO_TCP, TCP_NODELAY, (void *) &sock_parm, sizeof(sock_parm) );
      sock_parm = 1;
      setsockopt( newfd, SOL_SOCKET, SO_KEEPALIVE, (void *) &sock_parm, sizeof(sock_parm) );
      sock_parm = 16384;
      setsockopt( newfd, SOL_SOCKET, SO_SNDBUF, (void *) &sock_parm, sizeof(sock_parm) );
      sock_parm = 16384;
      setsockopt( newfd, SOL_SOCKET, SO_RCVBUF, (void *) &sock_parm, sizeof(sock_parm) );

      /*
       *  In case of troubles, the timeout prevents that a thread will block for ever.
       */
#if 0
      tv.tv_sec = 16 * 60;
      tv.tv_usec = 0;
      setsockopt( newfd, SOL_SOCKET, SO_RCVTIMEO, (void *) &tv, sizeof(tv) );
#endif
      tv.tv_sec = 1 * 30;	/* 30 seconds */
      tv.tv_usec = 0;
      setsockopt( newfd, SOL_SOCKET, SO_SNDTIMEO, (void *) &tv, sizeof(tv) );

      if (UFDBglobalDebug)
	 ufdbLogError( "SRV: accept new fd %2d", newfd );

      putFdQueue( newfd );
   }
}


/* TODO: update docs: to increase #server threads and #clients */
static void adjustNumberofThreads( void )
{
   int i;
   pthread_attr_t      attr;

   /* IF the check_https_tunnel_option is agressive we need more worker threads */
   if (UFDBgetTunnelCheckMethod() == UFDB_API_HTTPS_CHECK_AGRESSIVE  &&  n_workers < 32)
   {
      /* create more worker threads */
      i = n_workers;
      n_workers = 32;
      pthread_attr_init( &attr );
      pthread_attr_setscope( &attr, PTHREAD_SCOPE_SYSTEM );
      pthread_attr_setstacksize( &attr, 256 * 1024 );
      for (; i < n_workers; i++)
	 pthread_create( &workers[i], &attr, worker_main, (void *) ((long) i) );
      ufdbLogError( "#threads is increased to %d due to agressive proxy tunnel verifications", n_workers );
   }
}


void * accept_main( void * ptr )
{
   sigset_t            signals;
   int                 s;
   int                 sock_parm;
   struct sockaddr_in  addr;
   struct linger       linger;
   struct tms 	       timer;
   char                timeString[256];

   /* The HUP signal must be blocked.
    * This is a requirement to use sigwait() in a thread.
    */
   sigemptyset( &signals );
   sigaddset( &signals, SIGHUP );
   pthread_sigmask( SIG_BLOCK, &signals, NULL );

   UFDBtimerInit( &timer );
   gettimeofday( &start_time, NULL );

   globalArgv = my_argv;
   globalEnvp = my_envp;
   sgSetGlobalErrorLogFile();
   globalFatalError = 0;
   sgReadConfig( configFile );
   /* sgSetGlobalErrorLogFile(); */

   adjustNumberofThreads();

   if (noRedirects)
      ufdbLogError( "-T option is used.  In test mode no URLs are ever blocked." );
   if (globalFatalError)
      ufdbLogError( "A FATAL ERROR OCCURRED: ALL REQUESTS ARE ANSWERED WITH \"OK\"" );

   /*
    * create the daemon socket that the listen/accept connections on.
    */
   s = socket( AF_INET, SOCK_STREAM, 0 );
   if (s < 0)
   {
      ufdbLogError( "cannot create daemon socket: %s", strerror(errno) );
      pthread_exit( (void *) 1 );    /* TO-DO */
   }

   /* 
    * The port number can be specified in the config file (portNum is set by the parser)
    * or a command line parameter (portNumCmdLine is set in main).
    * The command line setting has preference.
    */
   if (portNumCmdLine >= 0)
      portNum = portNumCmdLine;
   addr.sin_family = AF_INET;
   addr.sin_addr.s_addr = htonl( INADDR_ANY );
   addr.sin_port = htons( portNum );

   /*
    * Allow server-side addresses to be reused (don't have to wait for timeout).
    */
   sock_parm = 1;
   setsockopt( s, SOL_SOCKET, SO_REUSEADDR, (void *) &sock_parm, sizeof(sock_parm) );

   if (bind( s, (struct sockaddr *) &addr, sizeof(addr) ) < 0)
   {
      ufdbLogError( "cannot bind daemon socket: %s", strerror(errno) );
      ufdbLogError( "check for and kill old daemon processes" );
      fprintf( stderr, "cannot bind daemon socket: %s\n", strerror(errno) );
      fprintf( stderr, "check for and kill old daemon processes" );
      close( s );
      pthread_exit( (void *) 1 );
   }
   
   /*
    * According to comment in the Apache httpd source code, these socket
    * options should only be set after a successful bind....
    *
    * Turn off NAGLE.
    */
   sock_parm = 1;
   setsockopt( s, SOL_SOCKET, SO_KEEPALIVE, (void *) &sock_parm, sizeof(sock_parm) );

   linger.l_onoff = 1;
   linger.l_linger = 2;
   setsockopt( s, SOL_SOCKET, SO_LINGER, (void *) &linger, sizeof(linger) );

   sock_parm = 1;
   setsockopt( s, IPPROTO_TCP, TCP_NODELAY, (void *) &sock_parm, sizeof(sock_parm) );

   if (listen( s, UFDB_MAX_WORKERS ) < 0)
   {
      ufdbLogError( "cannot listen on daemon socket: %s", strerror(errno) );
      fprintf( stderr, "cannot listen on daemon socket: %s\n", strerror(errno) );
      close( s );
      pthread_exit( (void *) 1 );
   }

   /* Now is the right moment to write out main PID to the pid file
    * that is used by /etc/init.d/ufdb to send signals to ufdbguardd.
    */
   writePidFile();
   atexit( removePidFile );

   ufdbLogError( "ufdbguardd " VERSION " started with %d verification threads", n_workers );

   UFDBreconfig = 0;

   daemon_accept_connections( s, (struct sockaddr *) &addr );

   UFDBtimerStop( &timer );
   UFDBtimerPrintString( timeString, &timer, "S" );
   ufdbLogError( timeString );

   removePidFile();
   pthread_exit( NULL );

   return NULL;
}


/*
 * the only task of this thread is to wait for the HUP signal.
 */
void * waitforsignals( void * ptr )
{
   sigset_t signals;
   int      sig;

   /* The HUP signal must be blocked.
    * This is a requirement to use sigwait() in a thread.
    */
   sigemptyset( &signals );
   sigaddset( &signals, SIGHUP );
   pthread_sigmask( SIG_BLOCK, &signals, NULL );

   ufdbSetThreadCPUaffinity();

   while (1)
   {
      sig = 0;
      if (0 != sigwait( &signals, &sig ))
         ufdbLogError( "sigwait() failed with %s", strerror(errno) );

      if (sig == SIGHUP)
      {
	 unsigned int sleep_period;

	 if (pthread_mutex_trylock( &sighup_mutex ) != 0)
	 {
	    ufdbLogError( "HUP signal is already being processed" );
	    continue;
	 }
         ufdbLogError( "HUP signal received" );

	 UFDBreconfig = 1;
	 if (analyseUncategorised)
	    saveUncategorisedURLs();

	 ufdbLogError( "statistics: %lu URL lookups (%lu https). %lu URLs blocked.", 
	               UFDBlookupCounter, 
		       UFDBlookupHttpsCounter,
		       UFDBblockCounter + UFDBtestBlockCounter );

	 /* sleep() is a very crude way to allow worker threads to finish what
	  * they were doing, before we remove the database from memory.
	  */
	 sleep_period = n_workers / 2;
	 if (sleep_period < 12)
	    sleep_period = 12;
	 sleep( sleep_period );		

         ufdbLogError( "freeing all memory..." );
	 ufdbFreeAllMemory();
         ufdbLogError( "reading new config..." );
	 globalFatalError = 0;
	 sgReadConfig( configFile );
	 if (globalFatalError)
	 {
	    ufdbLogError( "A FATAL ERROR OCCURRED: ALL REQUESTS ARE ANSWERED WITH \"OK\"" );
	 }
	 ufdbLogError( "the configuration is reloaded" );
	 adjustNumberofThreads();

	 UFDBlookupCounter = 0;
	 UFDBlookupHttpsCounter = 0;
	 UFDBblockCounter = 0;
	 UFDBtestBlockCounter = 0;
	 UFDBuncategorisedCounter = 0;
	 UFDBuploadSeqNo++;

	 UFDBreconfig = 0;
	 pthread_mutex_unlock( &sighup_mutex );
      }
      else 
         ufdbLogError( "signal %d received", sig );
   }
}


extern struct Destination * Dest;


/*
 * A pseudo-random sample of URLs is checked against all known categories
 * and if it appears to be uncategorised, it is buffered to be sent
 * to URLfilterDB at a later date (when SIGHUP is received).
 *
 * return 0 iff the URL is uncategorised, 1 otherwise.
 */
int ufdbVerifyURLallCats( UFDBrevURL * revURL )
{
   struct Destination *  cat;
   struct UFDBmemTable * mt;

   if (knownDb.table.nNextLevels > 0)
   {
      if (UFDBlookupRevUrl( &(knownDb.table.nextLevel[0]), revURL ))
         return 1;
   }

   for (cat = Dest; cat != NULL; cat = cat->next)
   {
      if (UFDBreconfig)
         return 0;
      if (cat->domainlistDb != NULL)
      {
	 mt = (struct UFDBmemTable *) cat->domainlistDb->dbcp;
	 if (UFDBlookupRevUrl( &(mt->table.nextLevel[0]), revURL ))
	    return 1;
      }
   }

   return 0;
}


static volatile int sample = 1;

/*
 * close a socket, read first all input.
 */
static void linger_close( void * fdp )
{
   int fd;
   char buffer[1024];

   fd = *( (int *) fdp);
   if (fd != -1)
   {
      struct timeval tv;

      tv.tv_sec = 1;
      tv.tv_usec = 0;
      setsockopt( fd, SOL_SOCKET, SO_RCVTIMEO, (void *) &tv, sizeof(tv) );
   }
   
   while (read( fd, buffer, 1024 ) == 1024)
      ;

   close( fd );
}


/*
 * main() of the worker threads.
 */
static void * worker_main( void * ptr )
{
   sigset_t                signals;
   UFDBthreadAdmin *       admin;
   int                     tnum;
   int                     fd;
   int                     nbytes;
   unsigned long           num_reqs;
   struct Source *         src;
   struct Acl *            acl;
   char * 	 	   redirect;
   struct SquidInfo        squidInfo;
   char *                  line;
   char *                  answer;
   char *                  tmp;
   char                    timeString[256];

   extern struct Source *  Source;

   tnum = (int) ((long) ptr);

   /* The HUP signal must be blocked.
    * This is a requirement to use sigwait() in a thread.
    */
   sigemptyset( &signals );
   sigaddset( &signals, SIGHUP );
   sigaddset( &signals, SIGPIPE );
   pthread_sigmask( SIG_BLOCK, &signals, NULL );

   ufdbSetThreadCPUaffinity();

   admin = UFDBallocThreadAdmin();

   line = ufdbCalloc( 1, MAX_BUF );
   answer = ufdbCalloc( 1, MAX_BUF );
   tmp = ufdbCalloc( 1, MAX_BUF );

   fd = -1;
   pthread_cleanup_push( linger_close, (void *) &fd );

   while (1)
   {
      struct tms timer;
      char       message[60];

      getFdQueue( &fd );
      UFDBtimerInit( &timer );
      num_reqs = 0UL;
      if (UFDBglobalDebug)
	 ufdbLogError( "W%02d: open fd %2d", tnum, fd );

      while ((nbytes = read( fd, line, MAX_BUF-2 )) > 0)
      {
         line[nbytes] = '\0';
         num_reqs++;
	 UFDBlookupCounter++;	/* no mutex: faster and we do not care about a few increments less */

         if (UFDBglobalDebug > 1)
	    ufdbLogError( "W%02d: line=%s", tnum, line );

	 if (globalFatalError || UFDBreconfig)
	 {
	    if (write( fd, "\n", 1 ) < 0)
	    {
	       ufdbLogError( "W%02d: write failed: %s", tnum, strerror(errno) );
	       goto write_error;
	    }
	    if (UFDBreconfig  &&  nbytes % 3 == 0)
	       sched_yield();	/* be nice to the reconfiguration thread */
	    continue;
	 }

	 squidInfo.revUrl = NULL;
	 if (parseLine(admin, line, &squidInfo) != 1) 	/* put input line in struct squidInfo */
	 {
	    ufdbLogError( "W%02d: error parsing input line: %s", tnum, line );
	    if (write( fd, "\n", 1 ) < 0)
	    {
	       ufdbLogError( "W%02d: write failed: %s", tnum, strerror(errno) );
	       UFDBfreeRevURL( admin, squidInfo.revUrl );
	       goto write_error;
	    }
	 }
	 else
	 {
#if 0
	    if (!UFDBreconfig  &&  UFDBlookupCounter % 23 == 0)
	       registerIP( squidInfo.src );				/* TODO */
#endif
            if (strcmp( squidInfo.protocol, "https" ) == 0)
	       UFDBlookupHttpsCounter++;

	    src = Source;
	    for (;;)
	    {
	       /* be extra careful with NOT using the in-memory database on a reload. */
	       if (UFDBreconfig) 
	       {
		  if (write( fd, "\n", 1 ) < 0)
		  {
		     ufdbLogError( "W%02d: write failed: %s", tnum, strerror(errno) );
		     UFDBfreeRevURL( admin, squidInfo.revUrl );
		     goto write_error;
		  }
		  break;
	       }

	       strcpy( tmp, squidInfo.src );	/* TODO: find out if strcpy() is necessary... */
	       src = sgFindSource( src, tmp, &squidInfo );

	       if (UFDBreconfig)
	       {
		  if (write( fd, "\n", 1 ) < 0)
		  {
		     ufdbLogError( "W%02d: write failed: %s", tnum, strerror(errno) );
		     UFDBfreeRevURL( admin, squidInfo.revUrl );
		     goto write_error;
		  }
		  break;
	       }
	       else
		  acl = sgAclCheckSource( src );

	       if (parseOnly || UFDBreconfig)
	       {
		  if (write( fd, "\n", 1 ) < 0)
		  {
		     ufdbLogError( "W%02d: write failed: %s", tnum, strerror(errno) );
		     UFDBfreeRevURL( admin, squidInfo.revUrl );
		     goto write_error;
		  }
		  break;
	       }

	       if ((redirect = sgAclAccess( src, acl, &squidInfo, tmp )) == NULL) 
	       {
		  /* src not found */
		  if (src == NULL || src->cont_search == 0) 
		  {
		     if (write( fd, "\n", 1 ) < 0)
		     {
			ufdbLogError( "W%02d: write failed: %s", tnum, strerror(errno) );
	                UFDBfreeRevURL( admin, squidInfo.revUrl );
			goto write_error;
		     }

		     if (analyseUncategorised && !UFDBreconfig)
		     {
	 		sample++;
			if (sample % 11 == 0)	/* approx. 9% is sampled */
			{
			   if (!ufdbVerifyURLallCats( squidInfo.revUrl ))
			   {
			      ufdbRegisterUnknownURL( squidInfo.domain );
			      UFDBuncategorisedCounter++;
			   }
			}
		     }

		     break;
		  }
		  else
		  {
		     if (!UFDBreconfig  &&  src->next != NULL) 
		     {
		        src = src->next;
		        continue;
		     }
		     else
		     {
		        if (write( fd, "\n", 1 ) < 0)
			{
			   ufdbLogError( "W%02d: write failed: %s", tnum, strerror(errno) );
			   UFDBfreeRevURL( admin, squidInfo.revUrl );
			   goto write_error;
			}
		        break;
		     }
		  }
	       }
	       else
	       {
		  /* src found, so block the URL and send the redirect string */
		  if (squidInfo.srcDomain[0] == '\0') 
		  {
		     squidInfo.srcDomain[0] = '-';
		     squidInfo.srcDomain[1] = '\0';
		  }

		  if (squidInfo.ident[0] == '\0') 
		  {
		     squidInfo.ident[0] = '-';
		     squidInfo.ident[1] = '\0';
		  }

		  if (globalDebugRedirect)
		  {
		      ufdbLogError( "W%02d:   REDIRECT %s %s/%s %s %s", tnum,
				    redirect, squidInfo.src,
				    squidInfo.srcDomain, squidInfo.ident, "GET" );
		  }

		  if (noRedirects)
		  {
		     char * category;

		     if (squidInfo.aclpass == NULL)
		        category = "error";
		     else if (squidInfo.aclpass->name == NULL)
		        category = "error";
		     else 
		        category = squidInfo.aclpass->name;
		     ufdbLogError( "TEST BLOCK %s %s", category, squidInfo.orig );
		     if (write( fd, "\n", 1 ) < 0)	/* send back an "OK" message */
		     {
			ufdbLogError( "W%02d: write failed: %s", tnum, strerror(errno) );
	                UFDBfreeRevURL( admin, squidInfo.revUrl );
			goto write_error;
		     }
		     UFDBtestBlockCounter++;
		  }
		  else
		  {
		     sprintf( answer, "%s %s/%s %s %s\n", redirect, squidInfo.src,
			      squidInfo.srcDomain, squidInfo.ident, "GET" );
		     if (write( fd, answer, strlen(answer) ) < 0)
		     {
			ufdbLogError( "W%02d: write failed: %s", tnum, strerror(errno) );
	                UFDBfreeRevURL( admin, squidInfo.revUrl );
			goto write_error;
		     }
		     UFDBblockCounter++;
		  }
		  break;
	       }
	    }
	    UFDBfreeRevURL( admin, squidInfo.revUrl );
	 }
	 if (sig_other)
	    break;
         pthread_testcancel();
      }
      if (nbytes < 0)
      {
         ufdbLogError( "W%02d: read failed: %s", tnum, strerror(errno) );
      }
write_error:

      UFDBtimerStop( &timer );

      if (num_reqs > 1UL)
      {
	 sprintf( message, "W%02d: %lu URL verifications in", tnum, num_reqs );
	 UFDBtimerPrintString( timeString, &timer, message );
	 ufdbLogError( timeString );
      }

      if (UFDBglobalDebug)
	 ufdbLogError( "W%02d: close fd %2d", tnum, fd );
      close( fd );
      fd = -1;
   }

   pthread_cleanup_pop( 0 );

   pthread_exit( NULL );
}


void ufdbCatchBadSignal( int signum )
{
   int       i;
   int       num;
   int       i_am_thread;
   pthread_t self;
   char      me[32];
   
   UFDBreconfig = 1;		/* be extra careful and set UFDBreconfig */

   self = pthread_self();

   /* find out which thread has a signal */
   i_am_thread = 0;
   sprintf( me, "pid %d", (int) getpid() );
   if (self == HUPhandler)
   {
      strcpy( me, "thread HUP-handler" );
      i_am_thread = 1;
   }
   else if (self == sockethandler)
   {
      strcpy( me, "thread socket-handler" );
      i_am_thread = 1;
   }
   else 
   {
      num = UFDB_NUM_HTTPS_VERIFIERS;
      for (i = 0; i < num; i++)
         if (self == httpsthr[i])
	 {
	    sprintf( me, "thread https-verifier-%02d", i );
	    i_am_thread = 1;
	    break;
	 }
      num = n_workers;
      for (i = 0; i < num; i++)
         if (self == workers[i])
	 {
	    sprintf( me, "thread worker-%02d", i );
	    i_am_thread = 1;
	    break;
	 }
   }
   ufdbLogError( "\"%s\" caught signal %d.", me, signum );

#if defined(__linux__) && 0
   if (signum != SIGINT && signum != SIGTERM)
   {
      void *    functions[32];
      char **   function_names;

      num = backtrace( functions, 32 );
      function_names = backtrace_symbols( functions, num );
      for (i = 0; i < num; i++)
	 ufdbLogError( "   function %s", function_names[i] );
      free( function_names );
   }
#endif

   if (signum == SIGINT || signum == SIGTERM)
   {
#if defined(UFDB_FREE_MEMORY)
      /* usleep( 200000 ); */
      sleep( 1 );
      ufdbLogError( "freeing memory ..." );
      ufdbFreeAllMemory();
      ufdbLogError( "done freeing memory." );
#endif
      /* implement our own linger_close() */
      removePidFile();
      pthread_cancel( sockethandler );
      sleep( 2 );
      exit( 0 );
   }
   else
   {
      removePidFile();

      if (i_am_thread)
      {
	 kill( globalPid, SIGABRT );
      }

      exit( 2 );
   }
}


static void catchHUPSignal( int signal )
{
   pthread_kill( HUPhandler, SIGHUP );
}


int main( 
   int            argc,
   char **        argv,
   char **        envp )
{
   int            i;
   pid_t          pid;
   pthread_attr_t attr;
   time_t 	  t;
   char           niso_str[22];

   strcpy( progname, "ufdbguardd" );

   while ((i = getopt(argc, argv, "ATdDPh?rt:c:L:p:vw:")) != EOF)
   {
      switch (i) {
      case 'A':
	 fprintf( stderr, "-A option is obsolete\n");
	 break;
      case 'N':
         analyseUncategorised = 0;
	 break;
      case 'D':
	 runAsDaemon = 0;	/* undocumented -D option to prevent running as daemon */
	 break;
      case 'd':
	 UFDBglobalDebug++;
	 break;
      case 'c':
	 configFile = optarg;
	 break;
      case 'p':
      	 portNumCmdLine = atoi( optarg );
	 if (portNumCmdLine <= 0) {
	    fprintf( stderr, "port number must be > 0\n" );
	    exit( 1 );
	 }
	 break;
      case 'P':			/* undocumented -P option for development purposes */
	 parseOnly = 1;
	 break;
      case 'r':
	 globalDebugRedirect = 1;
	 break;
      case 't':			/* undocumented -t option for time-related testing */
	 if ((t = iso2sec(optarg)) == -1) {
	    fprintf( stderr, "-t dateformat error, should be yyyy-mm-ddTHH:MM:SS\n" );
	    exit( 1 );
	 }
	 if (t < 0) {
	    fprintf( stderr, "-t date have to after 1970-01-01T01:00:00\n" );
	    exit( 1 );
	 }
	 niso( t, niso_str );
	 ufdbLogError( "ufdbguardd emulating date %s", niso_str );
	 globalDebugTimeDelta = t - start_time.tv_sec;
	 start_time.tv_sec = start_time.tv_sec + globalDebugTimeDelta;
	 break;
      case 'T':
	 noRedirects = 1;		/* Test mode means noRedirects */
	 break;
      case 'L':				/* undocumented -L option for alternate PID file */
	 globalPidFile = optarg;
         break;
      case 'v':
	 fprintf( stderr, "ufdbguardd: %s\n", VERSION );
	 exit( 0 );
	 break;
      case 'w':
         n_workers = atoi( optarg );
	 if (n_workers < 8)
	    n_workers = 8;
	 else if (n_workers > UFDB_MAX_WORKERS)
	    n_workers = UFDB_MAX_WORKERS;
	 break;
      case '?':
      case 'h':
         usage( '\0' );
	 break;
      default:
         usage( i );
      }
   }

   if (runAsDaemon)
   {
      if ((pid = fork()) != 0)
	 exit( 0 );        
      setsid();
   }

   globalPid = getpid();

   /*
    * Initialise signal handlers.
    * The HUP signal is dealt with on a per thread basis.
    */
   sgSetSignalHandler( SIGINT,  ufdbCatchBadSignal );
   sgSetSignalHandler( SIGQUIT, ufdbCatchBadSignal );
   sgSetSignalHandler( SIGILL,  ufdbCatchBadSignal );
   sgSetSignalHandler( SIGBUS,  ufdbCatchBadSignal );
   sgSetSignalHandler( SIGABRT, ufdbCatchBadSignal );
   sgSetSignalHandler( SIGSEGV, ufdbCatchBadSignal );
   sgSetSignalHandler( SIGTERM, ufdbCatchBadSignal );

   sgSetSignalHandler( SIGPIPE, SIG_IGN );
   sgSetSignalHandler( SIGHUP,  catchHUPSignal );

   initFdQueue();

   my_argc = argc;
   my_argv = argv;
   my_envp = envp;

   /*
    * create the threads.
    */
   pthread_attr_init( &attr );
   pthread_attr_setscope( &attr, PTHREAD_SCOPE_SYSTEM );
   pthread_attr_setstacksize( &attr, 256 * 1024 );

   pthread_create( &HUPhandler, &attr, waitforsignals, (void *) 0 );

   pthread_create( &sockethandler, &attr, accept_main, (void *) 0 );

   for (i = 0; i < UFDB_NUM_HTTPS_VERIFIERS; i++)
      pthread_create( &httpsthr[i], &attr, UFDBhttpsTunnelVerifier, (void *) ((long) i) );

   if (UFDBglobalDebug)
      ufdbLogError( "using %d worker threads", n_workers );
   for (i = 0; i < n_workers; i++)
      pthread_create( &workers[i], &attr, worker_main, (void *) ((long) i) );

   /*
    * exit if sockethandler thread dies.
    */
   pthread_join( sockethandler, 0 );

   /*
    * cleanup fast. cancel all other threads and wait a short while.
    */
   UFDBreconfig = 1;
   pthread_cancel( HUPhandler );
   for (i = 0; i < n_workers; i++)
   {
      pthread_cancel( workers[i] );
   }
   /* usleep( 100000 ); */
   sleep( 2 );
   exit( 0 );

#if 0
   for (i = 0; i < n_workers; i++)
      pthread_join( workers[i], 0);
#endif

   return 0;
}



static void saveUncategorisedURLs( void )
{
   char * URLs;
   char * message;
   int    length;
   int    s;
   int    nbytes;
   int    written;
   long   num_cpus;
   struct utsname sysinfo;

   if (UFDBglobalDebug)
      ufdbLogError( "saveUncategorisedURLs" );

   URLs = ufdbGetUnknownURLs();
   length = strlen( URLs );
   if (length == 0)
   {
      ufdbResetUnknownURLs();
      return;
   }

   message = ufdbMalloc( 1024 + length );
   if (message == NULL)
      return;

   s = UFDBopenSocket( UFDB_UPLOAD_UNCATEGORISED_URLS_WEBSITE, 80 );
   if (s < 0)
   {
      ufdbResetUnknownURLs();
      return;
   }

   if (uname( &sysinfo ) != 0)
   {
      strcpy( sysinfo.machine, "M?" );
      strcpy( sysinfo.release, "R?" );
   }

#if defined(_SC_NPROCESSORS_ONLN)
   num_cpus = sysconf( _SC_NPROCESSORS_ONLN );
   /* printf( "this system has %ld CPUs\n", num_cpus ); */

#elif defined(__NR_sched_getaffinity)
   /* sched_setaffinity() is buggy on linux 2.4.x so we use syscall() instead */
   cpu = syscall( __NR_sched_getaffinity, getpid(), 4, &cpu_mask );
   /* printf( "sched_getaffinity returned %d %08lx\n", cpu, cpu_mask ); */
   if (cpu >= 0)
   {
      num_cpus = 0;
      for (cpu = 0; cpu < 32; cpu++)
         if (cpu_mask & (1 << cpu))
            num_cpus++;
      /* printf( "   found %d CPUs in the cpu mask\n", num_cpus ); */
   }
   else
#endif
      num_cpus = 0;

   sprintf( message, 
	    "POST /cgi-bin/uncat.pl HTTP/1.1\r\n"
	    "Host: www.urlfilterdb.com\r\n"
	    "User-Agent: ufdbGuardd-" VERSION "\r\n"
	    "Content-Type: text/plain\r\n"
	    "Content-Length: %d\r\n"
	    "Connection: close\r\n"
	    "X-SiteInfo: %s %ld %s\r\n"
	    "X-NumLookups: %lu\r\n"
	    "X-NumHttpsLookups: %lu\r\n"
	    "X-NumTestBlock: %lu\r\n"
	    "X-NumBlock: %lu\r\n"
	    "X-Uncategorised: %lu\r\n"
	    "X-UploadSeqNo: %lu\r\n"
	    "\r\n"
	    "%s\r\n",
            length, 
	    sysinfo.machine, num_cpus, sysinfo.release, 
	    UFDBlookupCounter,
	    UFDBlookupHttpsCounter,
	    UFDBtestBlockCounter,
	    UFDBblockCounter,
	    UFDBuncategorisedCounter,
	    UFDBuploadSeqNo,
	    URLs );
   length = strlen( message );
   written = 0;
   while (length > 0)
   {
      nbytes = write( s, message+written, length );
      if (nbytes == 0)
         break;
      if (nbytes < 0)
      {
         if (errno != EINTR)
	    break;
      }
      else
      {
         written += nbytes;
	 length -= nbytes;
      }
   }

   close( s );

   ufdbFree( message );

   ufdbResetUnknownURLs();
   ufdbLogError( "uncategorised domains have been uploaded to URLfilterDB" );
}

